import logging
from distutils import config
from synch2jira.synch import Synchronisation
from synch2jira.issue_S1 import IssueImplementationS1
from synch2jira.issue_S2 import IssueImplementationS2

logging.basicConfig(filename=config.log_file, level=logging.INFO, format=config.log_format)
status_dict_S1_to_S2 = {"status1": "To Do", "status2": "Pret", "status3": "In Progress", "status4": "Done"}
status_dict_S1_to_S2 = {"En attente": "status1", "Pret": "status2", "En cours": "status3", "Qualifications": "status4"}


class ImplementationSynchonisationS1(Synchronisation):
    pass



from datetime import datetime 
import json
import requests
from requests.auth import HTTPBasicAuth

import config
from synch2jira.issue import Issue
from synch2jira.issue_S1 import IssueImplementationS1


class IssueImplementationS2(Issue):
    key: str

    def __init__(self, key=None, summary=None, description=None, updated=None, updated_status = None, status=None):
        Issue.__init__(self, summary, description, updated, status)
        self.key = key
        # if self.key is not None :
        #     self.miror = self.get_miror()  # id au lieu de IssueS1
        if self.miror is not None:
            self.miror.miror = self
        #self.workflow_history = self.get_worflow_history()

    def format_issue_into_json(self):
        issue = {
            "fields": {
                "description": {
                    "content": [
                        {
                            "content": [
                                {
                                    "text": self.description if self.description else "",
                                    "type": "text"
                                }
                            ],
                            "type": "paragraph"
                        }
                    ],
                    "type": "doc",
                    "version": 1
                },
                "issuetype": {
                    "id": config.key_issue_type
                },
                "project": {
                    "id": config.project_key
                },
                "labels":  config.labels if config.labels else [],
                "summary": self.summary
            },
            "update": {},
        }
        if self.miror is not None:
            issue['fields'][config.s1_id_in_jira] = str(self.miror.id)
        return json.dumps(issue)

    @staticmethod
    def convertir_json_en_issue(issue_json):
        try:
            issue_key = issue_json.get('key', None)
            issue_summary = issue_json['fields'].get('summary', None)
            description_data = issue_json['fields'].get('description', None)

            description_text = None
            if description_data is not None and 'content' in description_data:
                content_list = description_data['content']

                if content_list and content_list[0].get('content'):
                    description_text = content_list[0]['content'][0].get('text', None)

            issue_updated = issue_json['fields'].get('updated', None)
            status = issue_json['fields']['status']['name']
            return IssueImplementationS2(key=issue_key, summary=issue_summary, description=description_text,
                                         updated=issue_updated,
                                         status=status)
        except Exception as e:
            print(e)
            return None
    @classmethod
    def create_issue_lite(cls,issue_json):
        issue = IssueImplementationS2()
        issue.key = issue_json['key']
        issue.updated = issue_json['updated']
        return issue


    def get_miror(self):
        response = requests.request("GET", config.jira_url_ticket + self.key, headers=config.headers, auth=config.auth)
        if response.status_code == 200:
            mirror_id = response.json()['fields'].get(config.s1_id_in_jira, None)
            miror = IssueImplementationS1.find_by_id(mirror_id)
            return miror

    def get(self):
        try:
            response = requests.request("GET", config.jira_url_ticket + self.key, headers=config.headers, auth=config.auth)
            json_data = response.json()
            issue = IssueImplementationS2.convertir_json_en_issue(json_data)
            if issue is not None:
                return IssueImplementationS2(key=issue.key, summary=issue.summary, description=issue.description,
                                             updated=issue.updated, status=issue.status)
        except Exception as e:
            return e

    @staticmethod
    def find_by_id(key):
        try:
            response = requests.request("GET", config.jira_url_ticket + key, headers=config.headers, auth=config.auth)
            json_data = response.json()
            issue = IssueImplementationS2.convertir_json_en_issue(json_data)
            if issue is not None:
                return IssueImplementationS2(key=issue.key, summary=issue.summary, description=issue.description,
                                             updated=issue.updated, status=issue.status)
        except Exception as e:
            return e

    @staticmethod
    def find_by(jql_query):
        print(jql_query," ==== query")
        payload = {"jql": jql_query}
        response = requests.get(config.url_all, params=payload, headers=config.headers, auth=config.auth)           #à modifier 
        print("response code ",response.status_code)
        if response.status_code == 200:
            print("jql response",response.json().get("issues", []))
            return response.json().get("issues", [])
        else:
            return []

    # @staticmethod
    # def delete_all(jql_query):
    #     payload = {"jql": jql_query}
    #     response = requests.get(configurations.jira_url_all, params=payload, headers=configurations.headers, auth=configurations.auth)
    #     if response.status_code == 200:
    #         issues = response.json().get("issues", [])
    #         for issue in issues:
    #             IssueImplementationS2.delete(issue["key"])
    #     else:
    #         return f"Failed to execute search query. Status code: {response.status_code}"


    @staticmethod
    def get_jira_transitions_and_columns(issue_key):
        #transitions methode 2 
        transitions_url = f"{config.jira_url_base}/rest/api/2/issue/createmeta"
        # workflow_url = f"{configurations.jira_url_base}rest/api/2/workflow"
        # print("workflow url ",workflow_url)
        # #  la liste des workflows
        # workflows_response = requests.get(workflow_url, auth=configurations.auth)
        # print("repose",workflows_response)
        # workflows_data = workflows_response.json()
        # # on veux  les colonnes du premier workflow trouvé
        # if len(workflows_data) > 0:
        #     for work in workflows_data:
        #         print("work ...",work)
        #         if work["name"] == configurations.workflow_name:
        #             workflow_id = work["scope"]["project"]["id"]
        #             workflow_columns_url = f"{configurations.jira_url_base}rest/api/2/workflow/{workflow_id}/properties/jira.meta.statuses" #/properties/jira.meta.statuses
        #             print("workflow columnnn url ",workflow_columns_url)
        #             #  colonnes du workflow
        #             columns_response = requests.get(workflow_columns_url, auth=configurations.auth)
        #             print(columns_response)
        #             #columns_data = columns_response.json()
        #             #columns = [status['name'] for status in columns_data]
        #         else:
        #             columns = []
        return [],[]
    
    @staticmethod
    def get_jira_transitions(issue_key):
        transitions = []
        transition_url = config.jira_url_ticket + issue_key + "/transitions"
        transitions_response = requests.get(transition_url, auth=config.auth)
        transitions_data = transitions_response.json()
        transitions = [transition['name'] for transition in transitions_data['transitions']]
        return transitions

    def get_jira_status():
        status = [] 
        status_url = f"{config.jira_url_base}rest/api/3/status"
        status_response = requests.get(status_url, auth=config.auth)
        status_data = status_response.json()
        status = [status["name"] for status in status_data]
        return status

    
    # @staticmethod
    # def get_worflow_history(self):
    #     results = []
    #     history_url = f"{config.jira_url_ticket}{self.key}/changelog"
    #     history_response = requests.get(history_url, auth=config.auth)
    #     if history_response.status_code != 200:
    #         #print(f"Erreur lors de la récupération des transitions pour le ticket {issue_key}. Code d'erreur : {history_response.status_code}")
    #         return []
    #     i = 0
    #     history_data = history_response.json()
    #     histories = history_data['values']
    #     # print("last",histories[-1])
    #     # print("first",histories[0])
    #     # print(histories)
    #     for history in histories:
    #         created = history['created']
    #         items = history['items']
    #         for item in items:
    #             field = item['field']
    #             # print(field)
    #             # print(item)
    #             if field == "status":

    #                 print(f" history number {i+1}", history)
    #                 from_state = item['fromString']
    #                 to_state = item['toString']
    #                 #results.append([created,from_state,to_state])
    #                 history_dict = {'status_change_time': created, 'from_state': from_state, 'to_state':to_state}
    #                 results.append(history_dict)
    #                 #print("created at ",created,"from state ",from_state,"to state ",to_state)
    #                 i+=1
    #     #self.workflow_history = results
                    
    #     return results



    @staticmethod
    def all_():
        issues = IssueImplementationS2.get_issue_list_from_jira()
        issues_list = []
        for issue in issues:
            issues_list.append(IssueImplementationS2.convertir_json_en_issue(issue))

        return issues_list
    
    @staticmethod
    def all():
        if config.jql_query:
            return IssueImplementationS2.find_by(config.jql_query)
        issues = IssueImplementationS2.get_issue_list_from_jira()
        return issues
        
        #issue_list = IssueImplementationS2.find_by(config.jql_query)
        #return [IssueImplementationS2.convertir_json_en_issue(issue) for issue in issue_list]
    
    @staticmethod
    def all_lite():
        issues = IssueImplementationS2.get_issue_list_from_jira()
        #print(issues[0]['fields']['project'])
        return[IssueImplementationS2(key = issue["key"],updated=issue["fields"]["updated"]) for issue in issues]
    
    @staticmethod
    def all_weight():
        issues = IssueImplementationS2.get_issue_list_from_jira()
        print(issues,'issue wieght')
        return[IssueImplementationS2.convertir_json_en_issue(issue) for issue in issues]
    
    def get_issue_list_from_jira():
        response = requests.get(config.jira_url_all, headers=config.headers, auth=config.auth)
        issues = response.json()["issues"]
        return issues
    
    def all_jql():
        return[IssueImplementationS2(issue["key"],issue["fields"]["updated"]) for issue in IssueImplementationS2.find_by(config.jql_query)]


   
    
    @staticmethod
    def get_project_list():
        url = f'{config.url}/rest/api/3/project'
        response = requests.get(url, headers=config.headers, auth=config.auth1)
        
        if response.status_code == 200:
            projects = response.json()
 
            for project in projects:
                print(project)
                print(project['key'], '-', project['name'],'-',project['id'])
            return [[project['key'], project['name'],project['id']] for project in projects ]
        else:
            print(f"La requête a échoué avec le code d'erreur {response.status_code}")
        
    @staticmethod
    def all_filtre_id_et_updated():
        issues = IssueImplementationS2.get_issue_list_from_jira()
        issues_list = []
        for issue in issues:
            issue_key = issue.get('key', None)
            issue_updated = issue['fields'].get('updated', None)
            issues_list.append(
                IssueImplementationS2(key=issue_key, summary="", description="", updated=issue_updated, status=""))
        return issues_list

    def save(self):
        response = requests.request("POST", config.url_ticket, data=self.format_issue_into_json(), headers=config.headers,
                                    auth=config.auth1)    #à changer
        print(config.url_ticket)
        if response.status_code == 201:
            return (json.loads(response.text)).get("key")
        else:
            return f"Erreur lors de la création du ticket. Code d'erreur : {response.json()}"

    def delete(self):
        response = requests.request("DELETE", config.jira_url_ticket + self.key, auth=config.auth)

        if response.status_code == 204:
            print(f"Suppression effectuée avec succées")
        else:
            print(f"Erreur lors de la suppression du ticket. Code d'erreur : {response.status_code}")

    def update(self):
        response = requests.request("PUT", config.jira_url_ticket + self.key, data=self.format_issue_into_json(),
                                    headers=config.headers, auth=config.auth)
        return response.text

    def change_issue_status(self, new_status):
        transition_url = config.jira_url_ticket + self.key + "/transitions"
        response = requests.get(transition_url, auth=config.auth)
        transitions = response.json()['transitions']
        transition_id = None
        for transition in transitions:
            if transition['name'] == new_status:
                transition_id = transition['id']
                break
        if not transition_id:
            return None

        data = {
            "transition": {"id": transition_id}
        }
        headers = {'Content-Type': 'application/json'}

        response = requests.post(transition_url, json=data, headers=headers, auth=config.auth)

        if response.status_code == 204:
            return transition_id
        else:
            print("Erreur lors du changement de statut:", response.text)
            return None

    @staticmethod
    def supprimer_tickets_entre_a_and_b(a, b):
        for i in range(a, b):
            key = "KAN-" + str(i)
            issue_to_delete = IssueImplementationS2(key=key, summary="", description="", updated="", status="")
            issue_to_delete.delete()

    @staticmethod
    def check_connection(jira_url, username, api_token):
        try:
            auth = HTTPBasicAuth(username, api_token)
            response = requests.head(jira_url, auth=auth)
            if response.status_code == 200:
                return True
            else:
                return False
        except requests.RequestException as e:
            print("Erreur de connexion à Jira:", e)
            return False

    @staticmethod
    def try_jira_connection():
        response = requests.head(config.jira_url_base, auth=config.auth)
        if response.status_code == 200:
            return True
        return False




def config_package():
    with open('config.py', 'w') as fichier_config:
        fichier_config.write(f"""import logging
import os
from requests.auth import HTTPBasicAuth \n
username = "BillGates" 
#Api token de jira 
api_token = "TATT3xFfGF0nqgTV-RGN17B9CmizmQD0Mmr5ZY-pU0t8TjTzz0lyX0MNJ0XoNdKNy_t4eq9Is3Gw51Mta-kHF0XrEjKUANWzJM1XpRqS_-wSssC"

auth = HTTPBasicAuth(username, api_token)

jira_url_base = "https://example.atlassian.net/"
project_key = "90009"
key_issue_type = "10005"
s1_id_in_jira = "customfield_10054"

statusesS1 = ["status1", "status2", "status3", "status4"]
jiraStatusName = ["To Do", "Pret", "In Progress", "Done"]
jiraTransition = ["En attente", "Pret", "en cours", "Qualifications"]


module_to_use = "synch2jira.issue_S3"
class_to_use = "IssueImplementationS3"
""")




import os

main_directory = os.path.dirname(os.path.abspath(__file__))


log_file = main_directory + "/log/file.log"
log_format = '%(asctime)s - %(levelname)s - %(message)s'
table_correspondance_file = main_directory + "/files/correspondance_table.json"
history_file = main_directory + "/files/history.txt"


from sqlalchemy import Column, Integer, String, DateTime
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm import declarative_base
from synch2jira.issue import Issue
from sqlalchemy import create_engine
import config

db_url = config.database_file
Base = declarative_base()
engine = create_engine(db_url)
Session = sessionmaker(bind=engine)
session = Session()
Base.metadata.create_all(engine)


class IssueImplementationS3(Issue,Base):
    __tablename__ = 'issues'

    id = Column(Integer, primary_key=True, autoincrement=True)
    summary = Column(String)
    description = Column(String)
    updated = Column(DateTime)
    status = Column(String)
    miror = None

    def __init__(self, summary, description, updated, status):
        self.summary = summary
        self.description = description
        self.updated = updated
        self.status = status
    # def __repr__(self):
    #     return f"IssueImplementationS3(id={self.id}, summary={self.summary}, description={self.description}, updated={self.updated}, status={self.status})"

    @staticmethod
    def create(**kwargs):  # avec les argument on creee un objet et on le save
        issue = IssueImplementationS3(**kwargs)
        session.add(issue)
        session.commit()
        return issue

    @staticmethod
    def new(**kwargs):  # avec les argument on creee un objet sans le save
        return IssueImplementationS3(**kwargs)

    @staticmethod
    def all():
        return session.query(IssueImplementationS3).all()

    @staticmethod
    def first():
        return session.query(IssueImplementationS3).first()

    @staticmethod
    def last():
        return session.query(IssueImplementationS3).order_by(IssueImplementationS3.id.desc()).first()

    @staticmethod
    def find_by(**kwargs):
        return session.query(IssueImplementationS3).filter_by(**kwargs).all()

    @staticmethod
    def find_by_id(id):
        return session.query(IssueImplementationS3).filter_by(id=id).first()

    @staticmethod
    def save(instance):
        session.add(instance)
        session.commit()

    @staticmethod
    def update(instance, **kwargs):
        for key, value in kwargs.items():
            setattr(instance, key, value)
        session.commit()

    @staticmethod
    def update_all(**kwargs):
        session.query(IssueImplementationS3).update(kwargs)
        session.commit()

    @staticmethod
    def delete(instance):
        session.delete(instance)
        session.commit()

    @staticmethod
    def delete_all():
        session.query(IssueImplementationS3).delete()
        session.commit()





from dataclasses import dataclass
from datetime import datetime, timezone
from statistics import mean
from synch2jira.issue_S2 import IssueImplementationS2
from synch2jira.issue_wokflow import IssueWokflow

@dataclass
class WorkFlow():
    issueKey:str
    
    @staticmethod
    def get_duration_between_two_state(issue_key,state1,state2):
        if IssueWokflow.did_issue_have_state(issue_key,state1) and IssueWokflow.did_issue_have_state(issue_key,state2):
            worflow1 = IssueWokflow.find_by(issueKey = issue_key,status = state1)[0]
            worflow2 = IssueWokflow.find_by(issueKey = issue_key,status = state2)[0]
            date1 = WorkFlow.convert_str_time_to_unix(worflow1.from_time) 
            date2 = WorkFlow.convert_str_time_to_unix(worflow2.from_time)
            date3 = WorkFlow.convert_str_time_to_unix(worflow2.to_time)
            return [issue_key, worflow2.from_time,worflow1.from_time,date1-date2]
        else:
            pass

    @staticmethod 
    def get_all_duration_between_to_state(state1,state2,issue_list):            
        return [WorkFlow.get_duration_between_two_state(issue,state1,state2) for issue in issue_list]
    
    @staticmethod
    def get_all_workflow(state1,state2):
        issue_list = [issue.key for issue in IssueImplementationS2.all_lite()]
        workFlow_list = WorkFlow.get_all_duration_between_to_state(state1,state2,issue_list)
        return [workflow for workflow in workFlow_list if workflow ]

    
    @staticmethod 
    def get_mean_duration(state1,state2,issue_list):
        return mean(WorkFlow.get_duration_between_two_state(state1,state2,issue_list))
    
    @staticmethod  
    def is_in_state(state):
        return False
    
    @staticmethod
    def get_all_issues_with_state(cls,state):
        return []
    
    @staticmethod
    def get_rate(begin_time,end_time):
        workflow_list = IssueWokflow.find_by(status  = "Qualifications")
        print(f'before filtering {len(workflow_list)}')
        workflow_list = [workflow for workflow in workflow_list if workflow.to_time]
        print(f'after filtering {len(workflow_list)}')
        workflow_list = [workflow for workflow in workflow_list if WorkFlow.is_time_in_period(WorkFlow.convert_jira_date_to_datime(workflow.to_time),begin_time,end_time)]
        return len(workflow_list)

    @staticmethod
    def is_time_in_period(time,begin_time,end_time) :
        return begin_time <= time <= end_time
    
    @staticmethod
    def compare_date_with_str_date(date1,date2):
        date1 = date1.replace(tzinfo=timezone.utc)
        date2 = WorkFlow.convert_jira_date_to_datime(date2)
        print(f'date 1 {date1} , date 2 {date2}')
        if date1 > date2:
            #KAN-17 2024-03-05 09:13:40.998417 2024-03-05T11:27:28.959+0100
            #KAN-99 2024-03-05 11:06:19.485890 2024-03-05T11:27:29.733+0100
            return True
        
        return False 
    @staticmethod
    def convert_jira_date_to_datime(date):
        return datetime.strptime(date, "%Y-%m-%dT%H:%M:%S.%f%z").replace(tzinfo=timezone.utc)
    

    @staticmethod
    def get_worflow_history():
       pass

    @staticmethod
    def csv_export(cls,state1,state2):
        pass

    @staticmethod
    def convert_str_time_to_unix(time):
            timestamp_format = "%Y-%m-%dT%H:%M:%S.%f%z"
            element = datetime.strptime(time,timestamp_format)
            timestamp = datetime.timestamp(element)
            return timestamp
    


import logging
import os

import synch2jira.setup_files
from synch2jira.issue_S1 import IssueImplementationS1
from synch2jira.issue_S2 import IssueImplementationS2
from synch2jira.table_correpondance_key_id import verifier_existence_id, save_correspondance_table, \
    rebuild_correspondence_table, get_correspondence_table, \
    save_correspondence_dict

logging.basicConfig(filename=synch2jira.setup_files.log_file, level=logging.INFO, format=synch2jira.setup_files.log_format)


class Synchronisation:
    issue1 = IssueImplementationS1.issue1_factory(summary='', description='', updated=None, status='')

    @staticmethod
    def synch_S2_S1():
        if not (os.path.exists(synch2jira.setup_files.table_correspondance_file)) or os.path.getsize(
                synch2jira.setup_files.table_correspondance_file) == 0:
            print("try to rebuils correspondence table ")
            rebuild_correspondence_table()
        issues_in_S2 = IssueImplementationS2.all_filtre_id_et_updated()
        for issue_in_S2 in issues_in_S2:
            miror_S1_de_S2 = issue_in_S2.miror
            if miror_S1_de_S2 is None:
                issue_complet_S2 = issue_in_S2.get()
                issue_S2_convertie_en_S1 = IssueImplementationS1.convertirS_enS_(issue_complet_S2)
                issue_complet_S2.miror = issue_S2_convertie_en_S1.save()
                issue_complet_S2.update()
                logging.info(f"Ajout du ticket N°{issue_complet_S2.key} dans S1")

            # elif miror_S1_de_S2 and Issue.date_S1_entre_date_derniere_synchro_et_date_S2(miror_S1_de_S2.updated,
            #                                                                           issue_in_S2.updated):
            #     issue_complet_S2_convertie_en_S1.update()
            #     logging.info(f"Modification du ticket N°{issue_in_S2.key} dans S1")

            # issue_in_S1 = Issue.trouver(issue_in_S2, issues_in_S1)
            # issue_to_save_or_update = IssueImplementationS1.convertirS_enS_(issue_in_S2.get())
            # if issue_in_S1 is None:
            #     issue_to_save_or_update.save()
            #     logging.info(f"Ajout du ticket N°{issue_in_S2.key} dans S1")
            #     # Changer fonction date_S1_entre_date_derniere_synchro_et_date_S2 en utilisant mirror
            # elif issue_in_S1 and Issue.date_S1_entre_date_derniere_synchro_et_date_S2(issue_in_S1.updated,
            #                                                                           issue_in_S2.updated):
            #     issue_to_save_or_update.update()
            #     logging.info(f"Modification du ticket N°{issue_in_S2.key} dans S1")

    @staticmethod
    def synch_S1_S2():  # Créer a partir de S1 tout les S2
        issues_in_S1 = IssueImplementationS1.all_filtre_id_et_updated()
        for issue_in_S1 in issues_in_S1:
            if not verifier_existence_id(issue_in_S1.id):
                issue_complet_S1 = IssueImplementationS1.find_by_id(issue_in_S1.id)
                correspondance_S2 = IssueImplementationS2.convertirS_enS_(issue_complet_S1)
                correspondance_S2.miror = issue_in_S1
                key = correspondance_S2.save()
                save_correspondance_table(key, issue_in_S1.id)

    @staticmethod
    def synch_S1_S2_v2():
        if not (os.path.exists(synch2jira.setup_files.table_correspondance_file)) or os.path.getsize(
                synch2jira.setup_files.table_correspondance_file) == 0:
            print("try to rebuils correspondence table ... ")
            rebuild_correspondence_table()
            print("end of the construction ...")
        table = get_correspondence_table()
        table_id_list = table.values()
        issue = IssueImplementationS1.issue1_factory("", "", None, "")
        issue_list = issue.all()
        new_correspondences = {}
        for issue in issue_list:
            if str(issue.id) not in table_id_list:
                miror = IssueImplementationS2(summary=issue.summary, description=issue.description, status=issue.status)
                miror.miror = issue
                key = miror.save()
                if len(key) > 10:
                    print("error  ", key)
                    print("issue ====", issue)
                    print("idd ===", issue.id)
                    print("mirror ===", miror)
                    break
                new_correspondences[key] = str(issue.id)
        save_correspondence_dict(new_correspondences)


if __name__ == "__main__":
    Synchronisation.synch_S2_S1()
